"""Biosynthetic Gene Cluster prediction with Conditional Random Fields.

See Also:
    *Accurate de novo identification of biosynthetic gene clusters with GECCO*
    Laura M. Carroll, Martin Larralde, Jonas Simon Fleck, Ruby Ponnudurai,
    Alessio Milanese, Elisa Cappio, Georg Zeller. bioRxiv 2021.05.03.442509
    `doi:10.1101/2021.05.03.442509 <https://doi.org/10.1101/2021.05.03.442509>`_

"""

__author__ = "Martin Larralde"
__license__ = "GPLv3"
__version__ = "0.9.4"
